#!/usr/bin/perl

use strict;
use warnings;

use File::Basename;
use XML::Dumper (qw(pl2xml xml2pl));
use LWP::Simple (qw(get));

our $MySelf = defined ($ENV{'SCRIPT_NAME'}) ? $ENV{'SCRIPT_NAME'} : $0;
our $Config = read_config ();

if (defined ($ENV{'GATEWAY_INTERFACE'}))
{

	print <<EOH;
Content-Type: text/html
Cache-Control: no-cache

<html>
	<head>
		<title>Museum</title>
		<style type="text/css">
		img { display: block; border: none; }
		</style>
	</head>
	<body>
EOH

	action_html ($ENV{'PATH_INFO'} =~ m/([\w\-\.]+)/g);

	print <<EOF;
	</body>
</html>
EOF
}
else
{
	my @args = @ARGV;
	my $action;

	my $usage = <<EOU;
Usage: $0 <action> [opts]

Actions:
  add           Add collection
  del           Delete collection
  list          List collections
  ignore        Ignore a type or instance
  unignore      Unignore a type or instance
EOU

	if (!@args)
	{
		print STDOUT $usage;
		exit (1);
	}

	$action = shift (@args);

	if ($action eq 'add' or $action eq 'a')
	{
		action_add (@args);
	}
	elsif ($action eq 'del' or $action eq 'd')
	{
		action_del (@args);
	}
	elsif ($action eq 'list' or $action eq 'l')
	{
		action_list (@args);
	}
	elsif ($action eq 'ignore' or $action eq 'i')
	{
		action_ignore (@args);
	}
	elsif ($action eq 'unignore' or $action eq 'u')
	{
		action_unignore (@args);
	}
	elsif ($action eq 'html' or $action eq 'h')
	{
		action_html (@args);
	}
	else
	{
		print STDOUT $usage;
		exit (1);
	}
}

exit (0);

sub print_collections
{
	print "\t\t<h1>Welcome to my museum :)</h1>\n";
	print "\t\t<p>Which collection do you want to see?</p>\n";

	print "\t\t<ul>\n";
	for (sort (keys %{$Config->{'Collections'}}))
	{
		my $name = $_;

		print qq(\t\t\t<li><a href="$MySelf/$name">$name</a></li>\n);
	}
	print "\t\t</ul>\n";
}

sub print_host
{
	my $name = shift;

	die unless (defined ($Config->{'Collections'}{$name}));

	my $url = $Config->{'Collections'}{$name};

	print "\t\t<h1>Collection $name</h1>\n";
	print qq(\t\t<p><a href="$MySelf">Go back to my museum</a></p>\n);

	for (sort (keys %{$Config->{'Files'}{$name}}))
	{
		my $type = $_;
		for (@{$Config->{'Files'}{$name}{$type}})
		{
			my $inst = $_;

			my $imgurl = $url . '/graph/' . $type;
			$imgurl .= '/' . $inst if (length ("$inst"));
			$imgurl .= '/day';

			my $href = "$MySelf/$name/$type";
			$href .= "/$inst" if (length ("$inst"));

			next if ($Config->{'Ignore'}{$type}{$inst});
			
			print qq(\t\t<a href="$href"><img src="$imgurl" /></a>\n);
		}
	}
}

sub print_host_type
{
	my $name = shift;
	my $type = shift;
	my $inst = shift;

	$inst = '' unless (defined ($inst));

	die unless (defined ($Config->{'Collections'}{$name}));
	die unless (defined ($Config->{'Files'}{$name}{$type}));
	die unless (grep { "$_" eq "$inst" } (@{$Config->{'Files'}{$name}{$type}}));

	print "\t\t<h1>Collection $name</h1>\n";
	print "\t\t<h2>Type $type</h2>\n";
	print qq(\t\t<p><a href="$MySelf/$name">Go back to collection &quot;$name&quot;</a></p>\n);

	for (qw(day week month year))
	{
		my $span = $_;
		my $imgurl = $Config->{'Collections'}{$name} . '/graph/' . $type;
		$imgurl .= '/' . $inst if (length ("$inst"));
		$imgurl .= '/' . $span;

		print qq(\t\t<img src="$imgurl" />\n);
	}
}

sub action_html
{
	if (scalar (@_) > 3)
	{
		if (defined ($ENV{'GATEWAY_INTERFACE'}))
		{
			splice (@_, 3);
		}
		else
		{
			print <<EOU;
Usage: $0 html [name [name [inst]]] 
EOU
			exit (1);
		}
	}

	my $name = shift;
	my $type = shift;
	my $inst = shift;
	
	if (defined $type)
	{
		print_host_type ($name, $type, $inst);
	}
	elsif (defined $name)
	{
		print_host ($name);
	}
	else
	{
		print_collections ();
	}
}

sub action_add
{
	if (scalar (@_) != 2)
	{
		print <<EOU;
Usage: $0 add <name> <url>
EOU
		exit (1);
	}

	my ($name, $url) = @_;

	if (defined ($Config->{'Collections'}{$name}))
	{
		warn ("Error: A collection named ``$name'' exists. Delete it first.");
		exit (1);
	}

	if ($url =~ m#/$#)
	{
		$url .= 'collection.cgi';
	}
	
	my $xml = get ("$url/xml");
	my $info;

	if (!$xml)
	{
		warn ("Error: Unable to get data from $url");
		exit (1);
	}

	eval
	{
		$info = xml2pl ($xml);
	};
	if ($@)
	{
		warn ("Error: Could not parse XML: $@");
		exit (1);
	}

	$Config->{'Collections'}{$name} = $url;
	$Config->{'Files'}{$name} = $info;

	write_config ($Config);
}

sub action_del
{
	if (scalar (@_) != 1)
	{
		print <<EOU;
Usage: $0 del <name>
EOU
		exit (1);
	}

	my $name = shift;

	if (!defined ($Config->{'Collections'}{$name}))
	{
		warn ("No collection by the name ``$name''");
		exit (1);
	}

	delete ($Config->{'Collections'}{$name});
	delete ($Config->{'Files'}{$name});

	write_config ($Config);
}

sub action_list
{
	if (scalar (@_) > 1)
	{
		print <<EOU;
Usage: $0 list [name]
EOU
		exit (1);
	}

	my $name = @_ ? shift : '';

	if ($name)
	{
		if (!defined ($Config->{'Collections'}{$name}))
		{
			print STDERR "Collection ``$name'' could not be found.\n";
			exit (1);
		}

		print "Collection ``$name'' <" . $Config->{'Collections'}{$name} . ">\n";

		for (sort (keys %{$Config->{'Files'}{$name}}))
		{
			my $type = $_;
			my $ign = $Config->{'Ignore'}{$type}{''} ? ' (ignored)' : '';

			print "- $type$ign\n";

			for (@{$Config->{'Files'}{$name}{$type}})
			{
				next unless (length ("$_"));
				$ign = $Config->{'Ignore'}{$type}{$_} ? ' (ignored)' : '';
				print "  - $_$ign\n";
			}
		}
	}
	else
	{
		for (sort (keys %{$Config->{'Collections'}}))
		{
			my $name = $_;
			my $url = $Config->{'Collections'}{$name};

			print "- $name <$url>\n";
		}
	}
}

sub action_ignore
{
	if (scalar (@_) != 2 and scalar (@_) != 3)
	{
		print <<EOU;
Usage: $0 ignore <collection> <type> [instance]
EOU
		exit (1);
	}

	_action_ignore (1, @_);
}

sub action_unignore
{
	if (scalar (@_) != 2 and scalar (@_) != 3)
	{
		print <<EOU;
Usage: $0 unignore <collection> <type> [instance]
EOU
		exit (1);
	}

	_action_ignore (0, @_);
}

sub _action_ignore
{
	my $newv = shift;
	my $name = shift;
	my $type = shift;
	my $inst = @_ ? shift : undef;

	if (!defined ($Config->{'Collections'}))
	{
		warn ("No such collection ``$name''");
		exit (1);
	}
	if (!defined ($Config->{'Files'}{$name}{$type}))
	{
		warn ("No graph type ``$type'' in collection ``$name''");
		exit (1);
	}

	if (defined ($inst))
	{
		if (!grep { $_ eq $inst } (@{$Config->{'Files'}{$name}{$type}}))
		{
			warn ("No instance ``$inst'' of type ``$type'' in collection ``$name''");
			exit (1);
		}

		$Config->{'Ignore'}{$type}{$inst} = $newv;
	}
	else
	{
		for (@{$Config->{'Files'}{$name}{$type}})
		{
			$Config->{'Ignore'}{$type}{$_} = $newv;
		}
	}

	write_config ($Config);
}

sub read_config
{
	my $file = @_ ? shift : '/etc/museum.conf';
	my $conf;
	my $fh;

	if (open ($fh, "< $file"))
	{
		my $xml;
		local $/ = undef;
		$xml = <$fh>;

		eval
		{
			$conf = xml2pl ($xml);
		};
		warn ("Warning: eval xml2pl: $@") if ($@);
		close ($fh);
	}

	if (!$conf)
	{
		my $dir = dirname ($MySelf);
		my $host = defined ($ENV{'SERVER_NAME'}) ? $ENV{'SERVER_NAME'} : 'localhost';

		return ({
				Collections => {},
				Graphs => {}
			});
	}
	else
	{
		return ($conf);
	}
}

sub write_config
{
	my $config = shift;
	my $file = @_ ? shift : '/etc/museum.conf';
	my $fh;

	if (open ($fh, "> $file"))
	{
		print $fh pl2xml ($config);
		close ($fh);
	}
}

